using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using IntSight.Controls;
using IntSight.Controls.CodeModel;

namespace hammer
{
    /// <summary>
    /// An implementation of ICodeScanner for the Freya programming language.
    /// </summary>
    /// <remarks>
    /// This is the default scanner used by the code editor.
    /// When <c>CodeEditor.CodeScanner</c> contains a null pointer, the control
    /// uses an internal instance of this class.
    /// </remarks>
    internal sealed class FreyaScanner : ICodeScanner
    {
        public FreyaScanner() { }

        private sealed class KeySet : Dictionary<string, int>
        {
            public KeySet(params string[] keywords)
                : base(StringComparer.InvariantCultureIgnoreCase)
            {
                foreach (string s in keywords)
                    Add(s, 0);
            }

            /// <summary>Creates a set of rules for controlling block indentation.</summary>
            /// <param name="beginSet">Keywords that may start that block.</param>
            /// <param name="stopSet">Keywords that cannot belong to that block.</param>
            /// <param name="endSet">Keywords that close that block.</param>
            public KeySet(string[] beginSet, string[] stopSet, string[] endSet)
                : base(StringComparer.InvariantCultureIgnoreCase)
            {
                foreach (string s in beginSet)
                    Add(s, -1);
                foreach (string s in stopSet)
                    Add(s, 0);
                foreach (string s in endSet)
                    Add(s, +1);
            }

            public int Get(string token)
            {
                int value;
                if (TryGetValue(token, out value))
                    return value;
                else
                    return -2;
            }
        }

        private static KeySet keywords = new KeySet(
            "abstract", "and", "as", "begin", "break", "case", "class", "const",
            "constructor", "continue",
            "destructor", "div", "do", "downto", "else", "end", "ensures", "event",
            "except", "extern", "false", "fault", "finally", "for", "goto", "if",
            "implementation", "in", "interface", "internal", "invariant", "is", "iterator", "lock",
            "loop", "method", "mod", "namespace", "new", "nil", "not", "of",
            "old", "on", "operator", "or", "out", "override", "partial", "private",
            "property", "protected", "public", "raise", "readonly", "record", "repeat", "requires",
            "sealed", "shl", "shr", "static", "then", "to", "true", "try",
            "until", "using", "var", "virtual", "while", "yield", "xor");
        private static KeySet endKeySet = new KeySet(
            new string[] { "begin", "case", "except", "fault", "finally", "loop" } ,
            new string[] { "class", "constructor", "destructor", "event", "implementation",
                "interface", "internal" , "method", "namespace" , "private" , "property",
                "public", "static" },
            new string[] { "end" });
        private static KeySet finallyKeySet = new KeySet(
            new string[] { "try" } ,
            new string[] { "class", "constructor", "destructor", "event", "implementation",
                "interface", "internal", "method" , "namespace", "private" , "property" ,
                "public", "static", "except" },
            new string[] { "finally" });
        private static KeySet untilKeySet = new KeySet(
            new string[] { "repeat" } ,
            new string[] { "class", "constructor", "destructor", "event", "implementation",
                "interface", "internal" , "method", "namespace", "private" , "property",
                "public", "static" },
            new string[] { "until" });

        #region Precompiled regular expressions.

        private static Regex rxIndent = new Regex(
            "^(begin|else|except|finally|loop|implementation|internal|" +
            "private|protected|public|repeat|try|using|var|" +
            "internal (or|and) protected|protected (or|and) internal|" +
            "case .* of|for .* do|while .* do|if .* then|" +
            "using .* do|lock .* then|.*=\\s*interface|" +
            ".*=>|implementation for .* is)$",
            RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.IgnoreCase);
        private static Regex rxEndInner = new Regex(
            "^\\s*((method|property|constructor|event|iterator)(\\W|$)|end;?$)",
            RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.IgnoreCase);
        private static Regex rxEndMatch = new Regex(
            "^.*=\\s+(interface|record|class)(\\W|$)",
            RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.IgnoreCase);

        #endregion

        #region ICodeScanner members.

        void ICodeScanner.RegisterScanner(CodeEditor editor)
        {
            editor.DeleteTriggers();
            editor.AddTrigger("begin");
            editor.AddTrigger("else");
            editor.AddTrigger("end");
            editor.AddTrigger("ensures");
            editor.AddTrigger("finally");
            editor.AddTrigger("requires");
            editor.AddTrigger("until");
            editor.AddTrigger("public");
            editor.AddTrigger("private");
            editor.AddTrigger("implementation");
            editor.AddTrigger("protected");
            editor.AddTrigger("internal");
        }

        int ICodeScanner.DeltaIndent(
            CodeEditor codeEditor, string lastLine, int lastLineIndex)
        {
            if (rxIndent.IsMatch(lastLine))
                return +1;
            if (lastLineIndex == 0)
                return 0;
            int indent0, indent1;
            string token = codeEditor.GetFirstToken(lastLineIndex, out indent1);
            if (!keywords.ContainsKey(token))
            {
                token = codeEditor.GetFirstToken(lastLineIndex - 1, out indent0);
                if (indent0 + codeEditor.TabLength == indent1)
                    switch (token.ToLowerInvariant())
                    {
                        case "for":
                        case "while":
                        case "using":
                        case "if":
                        case "else":
                        case "lock":
                            return -1;
                    }
            }
            return 0;
        }

        private static char[] commentCharacters = { '/', '{', '}' };

        /// <summary>Does the supplied string contains any comment characters?</summary>
        /// <param name="text">String to be tested.</param>
        /// <returns>True if succeeds.</returns>
        bool ICodeScanner.ContainsCommentCharacters(string text)
        {
            return text.IndexOfAny(commentCharacters) != -1;
        }

        /// <summary>Is the given character part of a comment delimiter?</summary>
        /// <param name="ch">Character to be tested.</param>
        /// <returns>True if succeeds.</returns>
        bool ICodeScanner.IsCommentCharacter(char ch)
        {
            return ch == '{' || ch == '}' || ch == '/';
        }

        /// <summary>Gets the sequence that starts a line comment.</summary>
        string ICodeScanner.LineComment { get { return "//"; } }

        IEnumerable<Lexeme> ICodeScanner.Tokens(string text, bool comment)
        {
            StringBuilder sb = new StringBuilder();
            int p = 0, column = 0;
            bool lastWasExcept = false;
            char ch;
        START:
            if (p >= text.Length)
            {
                if (sb.Length > 0)
                    yield return new Lexeme(sb, ref column);
                yield break;
            }
            if (comment)
                ch = '{';
            else
                ch = text[p];
            switch (ch)
            {
                case '\n':
                case '\r':
                    {
                        p++;
                        if (ch == '\r' && p < text.Length && text[p] == '\n')
                            p++;
                        if (sb.Length > 0)
                            yield return new Lexeme(sb, ref column);
                        yield return Lexeme.NewLine;
                        column = 0;
                    }
                    goto START;
                case '$':
                    // An hexadecimal numeric literal.
                    {
                        int p0 = p;
                        do { p++; }
                        while (p < text.Length &&
                            "0123456789ABCDEFabcdef".IndexOf(text[p]) >= 0);
                        sb.Append(text, p0, p - p0);
                        lastWasExcept = false;
                    }
                    goto START;
                case '\'':
                case '"':
                    // A string literal.
                    // Freya accepts both double and single quoted string literals.
                    {
                        int p0 = p;
                        do { p++; }
                        while (p < text.Length && text[p] != ch &&
                            "\r\n".IndexOf(text[p]) < 0);
                        if (p < text.Length && text[p] == ch)
                            p++;
                        if (sb.Length > 0)
                            yield return new Lexeme(sb, ref column);
                        yield return new Lexeme(
                            Lexeme.Token.String, text.Substring(p0, p - p0), ref column);
                        lastWasExcept = false;
                    }
                    goto START;
                case '#':
                    // Delphi-like special characters in a string literal.
                    // Example: #13, #u0D, "Line 1"#13"Line 2".
                    {
                        int p0 = p;
                        p++;
                        if (p < text.Length && (text[p] == 'U' || text[p] == 'u'))
                        {
                            p++;
                            while (p < text.Length &&
                                "0123456789ABCDEFabcdef".IndexOf(text[p]) >= 0)
                                p++;
                        }
                        else
                            while (p < text.Length && Char.IsDigit(text, p))
                                p++;
                        if (sb.Length > 0)
                            yield return new Lexeme(sb, ref column);
                        yield return new Lexeme(
                            Lexeme.Token.String, text.Substring(p0, p - p0), ref column);
                        lastWasExcept = false;
                    }
                    goto START;
                case '/':
                    // A single line comment... or the plain operator for division.
                    {
                        int p0 = p;
                        p++;
                        if (p >= text.Length || text[p] != '/')
                        {
                            sb.Append("/");
                            lastWasExcept = false;
                            goto START;
                        }
                        do { p++; }
                        while (p < text.Length && "\r\n".IndexOf(text[p]) < 0);
                        if (sb.Length > 0)
                            yield return new Lexeme(sb, ref column);
                        yield return new Lexeme(
                            Lexeme.Token.Comment, text.Substring(p0, p - p0), ref column);
                    }
                    goto START;
                case '{':
                    // Multiline comments.
                    {
                        if (sb.Length > 0)
                            yield return new Lexeme(sb, ref column);
                        int p0 = p;
                        if (comment)
                        {
                            p--;
                            comment = false;
                        }
                        while (true)
                        {
                            p++;
                            if (p >= text.Length)
                            {
                                if (p > p0)
                                    yield return new Lexeme(Lexeme.Token.PartialComment,
                                        text.Substring(p0, p - p0), ref column);
                                yield break;
                            }
                            if (text[p] == '}')
                            {
                                p++;
                                break;
                            }
                            if (text[p] == '\n' || text[p] == '\r')
                            {
                                if (p > p0)
                                    yield return new Lexeme(Lexeme.Token.Comment,
                                        text.Substring(p0, p - p0), ref column);
                                yield return Lexeme.NewLine;
                                column = 0;
                                p++;
                                if (text[p - 1] == '\r')
                                    if (p < text.Length && text[p] == '\n')
                                        p++;
                                p0 = p;
                            }
                        }
                        if (p > p0)
                            yield return new Lexeme(Lexeme.Token.Comment,
                                text.Substring(p0, p - p0), ref column);
                    }
                    goto START;
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    // A numeric constant.
                    {
                        int p0 = p;
                        do p++; while (p < text.Length && Char.IsDigit(text, p));
                        sb.Append(text, p0, p - p0);
                        lastWasExcept = false;
                    }
                    goto START;
                case '}':
                    // A dangling end of comment.
                    {
                        p++;
                        if (sb.Length > 0)
                            yield return new Lexeme(sb, ref column);
                        yield return new Lexeme(Lexeme.Token.Error, "}", ref column);
                    }
                    goto START;
                default:
                    // Identifiers and keywords.
                    // The "AT" prefix is used with literal identifiers.
                    if (Char.IsLetter(ch) || ch == '_' || ch == '@')
                    {
                        int p0 = p;
                        do { p++; }
                        while (p < text.Length &&
                            (Char.IsLetterOrDigit(text, p) || text[p] == '_'));
                        string s = text.Substring(p0, p - p0);
                        if (keywords.ContainsKey(s))
                        {
                            if (sb.Length > 0)
                                yield return new Lexeme(sb, ref column);
                            if (String.Compare(s, "ON", true) == 0 && !lastWasExcept)
                                sb.Append(s);
                            else
                                yield return new Lexeme(Lexeme.Token.Keyword, s, ref column);
                            lastWasExcept = String.Compare(s, "EXCEPT", true) == 0;
                        }
                        else
                        {
                            sb.Append(s);
                            lastWasExcept = false;
                        }
                        goto START;
                    }
                    // Whitespace, what else?
                    if (Char.IsWhiteSpace(ch))
                    {
                        int p0 = p;
                        do { p++; }
                        while (p < text.Length && text[p] != '\r' &&
                            text[p] != '\n' && Char.IsWhiteSpace(text, p));
                        sb.Append(text, p0, p - p0);
                        goto START;
                    }
                    p++;
                    sb.Append(text, p - 1, 1);
                    lastWasExcept = false;
                    goto START;
            }
        }

        void ICodeScanner.TokenTyped(CodeEditor codeEditor, string token, int indent)
        {
            if (codeEditor.SmartIndentation &&
                codeEditor.CurrentColumn == indent + token.Length + 1)
                switch (token)
                {
                    case "begin":
                        UnindentBegin(codeEditor, indent);
                        break;
                    case "else":
                        UnindentElse(codeEditor, indent);
                        break;
                    case "end":
                        if (!Unindent(codeEditor, endKeySet, indent))
                            UnindentEnd(codeEditor, indent);
                        break;
                    case "requires":
                    case "ensures":
                        IndentAssertion(codeEditor, indent);
                        break;
                    case "finally":
                        Unindent(codeEditor, finallyKeySet, indent);
                        break;
                    case "until":
                        Unindent(codeEditor, untilKeySet, indent);
                        break;
                    case "public":
                    case "private":
                    case "protected":
                    case "implementation":
                    case "internal":
                        {
                            int ind, lineNo = codeEditor.CurrentLine - 1;
                            while (lineNo > 0)
                                switch (codeEditor.GetFirstToken(--lineNo, out ind))
                                {
                                    case "":
                                        break;
                                    case "method":
                                    case "property":
                                    case "constructor":
                                    case "iterator":
                                    case "event":
                                    case "end":
                                        if (ind == indent)
                                            Unindent(codeEditor, codeEditor.TabLength);
                                        return;
                                    default:
                                        return;
                                }
                        }
                        break;
                }
        }

        #endregion

        private static bool Unindent(CodeEditor codeEditor, KeySet keySet, int indent)
        {
            int lno = codeEditor.CurrentLine - 1, level = 0, ind;
            while (lno > 0)
                switch (keySet.Get(codeEditor.GetFirstToken(--lno, out ind)))
                {
                    case 0:
                        // We are out of the original token' natural scope.
                        return false;
                    case 1:
                        // We have reached the end of a nested scope.
                        level++;
                        break;
                    case -1:
                        if (level == 0)
                        {
                            Unindent(codeEditor, indent - ind);
                            return true;
                        }
                        else
                            // We have reached the beginning of a nested scope.
                            level--;
                        break;
                }
            return false;
        }

        private static void Unindent(CodeEditor codeEditor, int delta)
        {
            CodeEditor.Position p = codeEditor.Current;
            if (delta > 0 && p.Column > delta)
            {
                codeEditor.Select(p.Line, 0, delta);
                codeEditor.Delete();
                p = new CodeEditor.Position(p.Line, p.Column - delta);
                codeEditor.Select(p, p);
            }
        }

        private static void Indent(CodeEditor codeEditor, int delta)
        {
            CodeEditor.Position p = codeEditor.Current;
            codeEditor.Select(p.Line, 0, 0);
            codeEditor.SelectedText = new String(' ', delta);
            p.Column += delta;
            codeEditor.Select(p, p);
        }

        private static void IndentAssertion(CodeEditor codeEditor, int indent)
        {
            if (codeEditor.CurrentLine > 1)
            {
                int ind;
                string token = codeEditor.GetFirstToken(codeEditor.CurrentLine - 2, out ind);
                if (!String.IsNullOrEmpty(token))
                    switch (token.ToLowerInvariant())
                    {
                        case "constructor":
                        case "iterator":
                        case "method":
                        case "operator":
                        case "property":
                            {
                                int delta = indent - ind - codeEditor.TabLength;
                                if (delta > 0)
                                    Unindent(codeEditor, delta);
                                else if (delta < 0)
                                    Indent(codeEditor, -delta);
                            }
                            break;
                    }
            }
        }

        private static void UnindentElse(CodeEditor codeEditor, int indent)
        {
            if (codeEditor.CurrentLine > 1)
            {
                int ind;
                string token = codeEditor.GetFirstToken(codeEditor.CurrentLine - 2, out ind);
                if (!String.IsNullOrEmpty(token))
                    switch (token.ToLowerInvariant())
                    {
                        case "end":
                        case "if":
                            if (ind < indent)
                                Unindent(codeEditor, indent - ind);
                            break;
                        default:
                            if (ind >= codeEditor.TabLength &&
                                indent > ind - codeEditor.TabLength)
                                Unindent(codeEditor, indent - ind + codeEditor.TabLength);
                            break;
                    }
            }
        }

        private static void UnindentBegin(CodeEditor codeEditor, int indent)
        {
            if (codeEditor.CurrentLine < 2)
                return;
            string line = codeEditor[codeEditor.CurrentLine - 2];
            if (String.IsNullOrEmpty(line))
                return;
            int ind;
            string token = codeEditor.GetFirstToken(codeEditor.CurrentLine - 2, out ind);
            if (ind == indent - codeEditor.TabLength)
                switch (token.ToLowerInvariant())
                {
                    case "for":
                    case "while":
                    case "using":
                        if (line.EndsWith(" do", StringComparison.InvariantCultureIgnoreCase))
                            Unindent(codeEditor, codeEditor.TabLength);
                        break;
                    case "lock":
                    case "if":
                        if (line.EndsWith(" then", StringComparison.InvariantCultureIgnoreCase))
                            Unindent(codeEditor, codeEditor.TabLength);
                        break;
                    case "else":
                        Unindent(codeEditor, codeEditor.TabLength);
                        break;
                }
        }

        private static void UnindentEnd(CodeEditor codeEditor, int indent)
        {
            int lNo = codeEditor.CurrentLine - 1;
            while (lNo-- > 0)
            {
                string line = codeEditor[lNo];
                if (!String.IsNullOrEmpty(line))
                {
                    if (rxEndInner.IsMatch(line))
                    {
                        int ind = codeEditor.GetIndentation(lNo) - codeEditor.TabLength;
                        if (ind >= 0)
                            if (ind < indent)
                                Unindent(codeEditor, indent - ind);
                            else if (ind > indent)
                                Indent(codeEditor, ind - indent);
                        return;
                    }
                    if (rxEndMatch.IsMatch(line))
                    {
                        int ind = codeEditor.GetIndentation(lNo);
                        if (ind < indent)
                            Unindent(codeEditor, indent - ind);
                        else if (ind > indent)
                            Indent(codeEditor, ind - indent);
                        return;
                    }
                }
            }
        }
    }
}
